#include "syshdrs.h"

#define _CRT_SECURE_NO_WARNINGS 1

#ifdef PRAGMA_HDRSTOP
#	pragma hdrstop
#endif

#ifndef HAVE_SOCKADDR_SA_LEN
#ifndef SA_LEN
#ifdef ENABLE_IPV6
#define SA_LEN(x) (((x)->sa_family == AF_INET6) ? sizeof(struct sockaddr_in6) \
		                : (((x)->sa_family == AF_INET) ? sizeof(struct sockaddr_in) \
					                : -1))
#else  /* ENABLE_IPV6 */
#define SA_LEN(x) (((x)->sa_family == AF_INET) ? sizeof(struct sockaddr_in) : -1)
#endif /* ENABLE_IPV6 */
#endif /* SA_LEN */
#endif /* HAVE_SOCKADDR_SA_LEN */


#ifndef INADDR_ANY
#	define INADDR_ANY              ((unsigned long int) 0x00000000)
#endif

#ifndef INADDR_NONE
#	define INADDR_NONE             ((unsigned long int) 0xffffffff)
#endif

/* Linux libc 5.3.x has a bug that causes isalnum() to not work! */
#define ISALNUM(c) ( (((c) >= 'A') && ((c) <= 'Z')) || (((c) >= 'a') && ((c) <= 'z')) || (((c) >= '0') && ((c) <= '9')) )

#if (defined(WIN32) || defined(_WINDOWS)) && !defined(__CYGWIN__)
#	pragma warning(disable : 4711)	// warning: function 'ServiceNameToPortNumber' selected for automatic inline expansion
#endif
unsigned int
ServiceNameToPortNumber(const char *const s, const int proto)
{
	char str[64];
	char *cp;
	struct addrinfo hints, *res;
	int error;
	unsigned int port;

	strncpy(str, s, sizeof(str) - 1);
	str[sizeof(str) - 1] = '\0';
	cp = str;
	if (isdigit((int) *cp)) {
		while (isdigit((int) *cp))
			cp++;
		*cp = '\0';
		return (atoi(str));
	}
	for (;; cp++) {
		if ((*cp == '\0')
			|| ((!ISALNUM(*cp)) && (*cp != '-') && (*cp != '_')))
				break;
	}
	*cp = '\0';

	error = 1;
	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	if ((error != 0) && ((proto == 0) || (proto == 't'))) {
		hints.ai_socktype = SOCK_STREAM;
		error = getaddrinfo(NULL, str, &hints, &res);
	}
	if ((error != 0) && ((proto == 0) || (proto == 'u'))) {
		hints.ai_socktype = SOCK_DGRAM;
		error = getaddrinfo(NULL, str, &hints, &res);
	}
	if (error != 0) {
		return (0);	/* error */
	}
	if (res->ai_addr->sa_family == AF_INET) {
		port = ((struct sockaddr_in *)res->ai_addr)->sin_port;
	} else if (res->ai_addr->sa_family == AF_INET6) {
		port = ((struct sockaddr_in6 *)res->ai_addr)->sin6_port;
	} else {
		port = 0;	/* error */
	}
	freeaddrinfo(res);
	return ((unsigned int) ntohs((unsigned short) port));
}	/* ServiceNameToPortNumber */




int
ServicePortNumberToName(unsigned short port, char *const dst, const size_t dsize, const int proto)
{
#if defined(HAVE_GETSERVBYNAME_R) && (defined(AIX) || defined(TRU64UNIX) || defined(DIGITAL_UNIX))
	struct servent *sp;
#elif defined(HAVE_GETSERVBYNAME_R) && (defined(LINUX) || defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
	struct servent se, *sp;
	char spbuf[256];
#else
	struct servent *sp;
#endif

	sp = NULL;
#if defined(HAVE_GETSERVBYPORT_R) && (defined(SOLARIS) || defined(IRIX) || defined(BSDOS))
	if ((sp == NULL) && ((proto == 0) || (proto == 't'))) {
		memset(spbuf, 0, sizeof(spbuf));
		sp = getservbyport_r((int) htons(port), "tcp", &se, spbuf, sizeof(spbuf));
	}
	if ((sp == NULL) && ((proto == 0) || (proto == 'u'))) {
		memset(spbuf, 0, sizeof(spbuf));
		sp = getservbyport_r((int) htons(port), "udp", &se, spbuf, sizeof(spbuf));
	}
#elif defined(HAVE_GETSERVBYPORT_R) && defined(LINUX)
	if ((sp == NULL) && ((proto == 0) || (proto == 't'))) {
		memset(spbuf, 0, sizeof(spbuf));
		if (getservbyport_r((int) htons(port), "tcp", &se, spbuf, sizeof(spbuf), &sp) != 0)
			sp = NULL;
	}
	if ((sp == NULL) && ((proto == 0) || (proto == 'u'))) {
		memset(spbuf, 0, sizeof(spbuf));
		if (getservbyport_r((int) htons(port), "udp", &se, spbuf, sizeof(spbuf), &sp) != 0)
			sp = NULL;
	}
#elif defined(HAVE_GETSERVBYPORT_R) && defined(AIX)
	{
		struct servent_data sed;
		if ((sp == NULL) && ((proto == 0) || (proto == 't'))) {
			memset(&sed, 0, sizeof(sed));
			if (getservbyport_r((int) htons(port), "tcp", sp, &sed) != 0)
				sp = NULL;
		}
		if ((sp == NULL) && ((proto == 0) || (proto == 'u'))) {
			memset(&sed, 0, sizeof(sed));
			if (getservbyport_r((int) htons(port), "udp", sp, &sed) != 0)
				sp = NULL;
		}
	}
#else
	/* Note: getservbyport is already threadsafe on: HP-UX, Tru64 */
	if ((sp == NULL) && ((proto == 0) || (proto == 't'))) {
		sp = getservbyport((int) htons(port), "tcp");
	}
	if ((sp == NULL) && ((proto == 0) || (proto == 'u'))) {
		sp = getservbyport((int) htons(port), "ucp");
	}
#endif

	if (sp != NULL) {
		strncpy(dst, sp->s_name, dsize);
		dst[dsize - 1] = '\0';
		return (1);
	}

#ifdef HAVE_SNPRINTF
	snprintf(dst, dsize,
#else
	sprintf(dst,
#endif
		"%u", (unsigned int) port);

	return (0);	/* error */
}	/* ServicePortNumberToName */




void
InetNtoA(char *dst, struct in_addr *ia, size_t siz)
{
#if defined(HAVE_INET_NTOP) && !defined(MACOSX)
	/* Mostly to workaround bug in IRIX 6.5's inet_ntoa */
	/* For OS X, don't use inet_ntop yet since it was just introduced
	 * for 10.2.
	 */
	memset(dst, 0, siz);
	(void) inet_ntop(AF_INET, ia, dst, siz - 1);
#else
	char *cp;
	memset(dst, 0, siz);
	cp = inet_ntoa(*ia);
	if ((cp != (char *) 0) && (cp != (char *) -1) && (cp[0] != '\0')) {
		(void) strncpy(dst, cp, siz - 1);
	}
#endif
}	/* InetNtoA */




int
AddrStrToAddr(const char * const s, struct sockaddr * const sa, const int defaultport)
{
	char portstr[128], portstr2[128];
	struct addrinfo hints, *res;
	char *hostcp, *atsign, *colon, *cp, *p2, *portp;
	int error;

	strncpy(portstr, s, sizeof(portstr));
	portstr[sizeof(portstr) - 1] = '\0';

	if ((colon = strchr(portstr, ':')) != NULL) {
		/* Does it look like a URL?  http://host ? */
		if ((colon[1] == '/') && (colon[2] == '/')) {
			*colon = '\0';
			hostcp = colon + 3;
			for (cp = hostcp; *cp != '\0'; cp++) {
				if ((!ISALNUM(*cp)) && (*cp != '.')) {
					/* http://host:port */
					if ((*cp == ':') && (isdigit((int) cp[1]))) {
						*cp++ = '\0';
						p2 = cp;
						while (isdigit((int) *cp))
							cp++;
						*cp = '\0';
					}
					*cp = '\0';
					break;
				}
			}
		} else {
			/* Look for host.name.domain:port */
			*colon = '\0';
			hostcp = portstr;
		}
		portp = portstr;
	} else if ((atsign = strchr(portstr, '@')) != NULL) {
		/* Look for port@host.name.domain */
		*atsign = '\0';
		hostcp = atsign + 1;
		portp = portstr;
	} else if (defaultport > 0) {
		/* Have just host.name.domain, use that w/ default port. */
		sprintf(portstr2, "%d", defaultport);
		hostcp = portstr;
		portp = portstr2;
	} else {
		/* If defaultport <= 0, they must supply a port number
		 * in the host/port string.
		 */
		errno = EADDRNOTAVAIL;
		return (kAddrStrToAddrMiscErr);
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = PF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	error = getaddrinfo(hostcp, portp, &hints, &res);
	if (error != 0)
		return (kAddrStrToAddrBadHost);
	memcpy(sa, res->ai_addr, res->ai_addrlen);
	freeaddrinfo(res);
	return (0);
}	/* AddrStrToAddr */




char *
AddrToAddrStr(char *const dst, size_t dsize, struct sockaddr * const saddrp, int dns, const char *fmt)
{
	char addrName[128];
	char *addrNamePtr;
	char portName[128];
	char portNum[128];
	char s_name[64];
	char *dlim, *dp;
	const char *cp;

	if (dsize == 0)
		return NULL;
	memset(dst, 0, dsize);

	addrNamePtr = addrName;
	if (dns == 0)
		dns = NI_NUMERICHOST;
	else
		dns = 0;
	getnameinfo(saddrp,
#ifdef HAVE_SOCKADDR_SA_LEN
		    saddrp->sa_len,
#else
		    SA_LEN(saddrp),
#endif
		    addrName, sizeof(addrName), portName, sizeof(portName), dns);
	getnameinfo(saddrp,
#ifdef HAVE_SOCKADDR_SA_LEN
		    saddrp->sa_len,
#else
		    SA_LEN(saddrp),
#endif
		    NULL, 0, portNum, sizeof(portNum), NI_NUMERICSERV);
	if (fmt == NULL)
		fmt = "%h:%p";
	for (dp = dst, dlim = dp + dsize - 1; ; fmt++) {
		if (*fmt == '\0') {
			break;	/* done */
		} else if (*fmt == '%') {
			fmt++;
			if (*fmt == '%') {
				if (dp < dlim)
					*dp++ = '%';
			} else if (*fmt == 'p') {
				for (cp = portNum; *cp != '\0'; cp++)
					if (dp < dlim)
						*dp++ = *cp;
				*dp = '\0';
			} else if (*fmt == 'h') {
				if (addrNamePtr != NULL) {
					cp = addrNamePtr;
				} else {
					cp = "unknown";
				}
				for ( ; *cp != '\0'; cp++)
					if (dp < dlim)
						*dp++ = *cp;
				*dp = '\0';
			} else if (*fmt == 's') {
				for (cp = portName ; *cp != '\0'; cp++)
					if (dp < dlim)
						*dp++ = *cp;
				/* endservent(); */
				*dp = '\0';
			} else if ((*fmt == 't') || (*fmt == 'u')) {
				cp = s_name;
				if (saddrp->sa_family == AF_INET) {
					(void) ServicePortNumberToName(ntohs(((struct sockaddr_in *)saddrp)->sin_port), s_name, sizeof(s_name), (int) *fmt);
				} else if (saddrp->sa_family == AF_INET6) {
					(void) ServicePortNumberToName(ntohs(((struct sockaddr_in6 *)saddrp)->sin6_port), s_name, sizeof(s_name), (int) *fmt);
				} else {
					cp = '\0';
				}
				for ( ; *cp != '\0'; cp++)
					if (dp < dlim)
						*dp++ = *cp;
				/* endservent(); */
				*dp = '\0';
			} else if (*fmt == '\0') {
				break;
			} else {
				if (dp < dlim)
					*dp++ = *fmt;
			}
		} else if (dp < dlim) {
			*dp++ = *fmt;
		}
	}
	*dp = '\0';
	return (dst);
}	/* AddrToAddrStr */




char *
AddrStrToIPStr(char *const dst, size_t dsize, const char *const src, const int defaultport)
{
	int rc;
	struct sockaddr_storage ss;

	if (dsize == 0)
		return NULL;
	memset(dst, 0, dsize);

	rc = AddrStrToAddr(src, (struct sockaddr *)&ss, (defaultport <= 0) ? 21 : defaultport);
	if (rc < 0)
		return (NULL);

	AddrToAddrStr(dst, dsize, (struct sockaddr *)&ss, 0, (defaultport <= 0) ? "%h" : "%h:%p");
	return (dst);
}	/* AddrStrToIPStr */
